/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Profile Manager.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Jonathan Griffin <jgriffin@mozilla.com>
 *    Mark Cote <mcote@mozilla.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// a list of symbols to export; other files can import these symbols
// using Components.utils.import(..)
var EXPORTED_SYMBOLS = ["ProfileManagerService"];

const Ci = Components.interfaces;
const Cc = Components.classes;

// XPCOM services
var gDirectoryService = Cc["@mozilla.org/file/directory_service;1"].
                        getService(Ci.nsIProperties);

function thunderbirdProfileDirectory() {
  if (gPlatform == "X11") {
    // linux stores thunderbird profiles in ~/.thunderbird

    let profileDirectory = gDirectoryService.get("Home", Ci.nsIFile);
    profileDirectory.append('.thunderbird');
    return profileDirectory;
  } else {
    let profileDirectory = getFile(getEnv('APPDATA'));
    profileDirectory.append('Thunderbird');
    return profileDirectory;
  }
}

function ExtraOptionBool(name, locale_desc, option_value) {
  // Create an extra option for a boolean value.
  // If other types of values are required, we can create a base class.
  // Parameter 'locale_desc' refers to the localization property name
  //  containing the description for the option--see
  //  chrome/locale/*/profileManager.properties

  this.name = name;
  this.desc = document.getElementById("bundle_profileManager").getString(locale_desc);
  this.option_value = option_value;
  this.opttype = "bool";
  this.option = function() {
    if (this.selected())
      return this.option_value;
    return "";
  };

  // Set by UI code.
  this.selected = function() { return false; };
}

function firefoxExtraOptions() {
  return [new ExtraOptionBool("offline",
			      "ffxOptOfflineModeDesc",
			      "-offline")];
}

var ProfileManagerService = 
{
  //
  // define 'private' variables
  //

  // the app's short name (e.g., firefox)
  _app: undefined,
  // the app's title (e.g., Firefox)
  _appName: undefined,
  // nsIFile of the profile directory
  _profileDirectory: undefined,
  // JS object represesnting the default app, with 'path' and 'version'
  // properties
  _defaultApp: undefined,
  // application to full-name mapping
  // Each mapping may have the following properties:
  //    name: the display name for the app
  //    default_subdir: the subdirectory under "Program Files"
  //           where the app is installed by default on Windows
  //    profile_directory: callable which returns the path to profiles.ini
  //           for the application
  //    extra_opts: array of extra command-line options, if any
  _appsData: {firefox:      {name: 'Firefox',
                            default_subdir: 'Mozilla Firefox',
                            extra_opts: firefoxExtraOptions()},
             thunderbird:  {name: 'Thunderbird',
                            default_subdir: 'Mozilla Thunderbird',
                            profile_directory: thunderbirdProfileDirectory}
  },

  //
  // define 'public' variables
  //
  profileManager: undefined,
  appConfig: undefined,

  //
  // define getters and setters for various properties
  //
  get appProperties() {
    if (this._app)
      return this._appsData[this._app];
  },

  get defaultApp() {
    return this._defaultApp;
  },
  get defaultAppPath() {
    if (this._defaultApp) {
      return this._defaultApp.path;
    }
    else {
      return '';
    }
  },
  get extraOpts() {
    if (this._appsData[this._app].extra_opts) {
      return this._appsData[this._app].extra_opts;
    }
    else {
      return [];
    }
  },
  set defaultApp(app) {
    if (typeof(app) == "object") {
      if (app instanceof Ci.nsIFile)
        this._defaultApp = { 'path':app.path, 'version':getAppVersion(app.path) };
      else
        this._defaultApp = copyObject(app);
    }
    else if (typeof(app) == "string")
      this._defaultApp = { 'path':app, 'version':getAppVersion(app) };
    else
      this._defaultApp = undefined;
  },

  get defaultSubdir() {
    if (this._appsData[this._app]) {
      return this._appsData[this._app].default_subdir;
    }
  },

  get appTitle() {
    if (this._appsData[this._app]) {
      return this._appsData[this._app].name;
    }
  },

  set profileDirectory(dir) {
    this._profileDirectory = dir;
  },
  get profileDirectory() {
    return this._profileDirectory;
  },

  get appName() {
    return this._appName;
  },

  set app(app) {
    this._app = app;

    // set app name
    if ( this._appsData[this._app] != undefined) {
      this._appName = this._appsData[this._app].name;
    }
    if ( this._appName == undefined ) {
      this._appName = this._app;
    }

    // set profile directory
    if (this._appsData[this._app] && 
        this._appsData[this._app].profile_directory) {
      // app specific profile directory
      this._profileDirectory = this._appsData[this._app].profile_directory();
    }
    else {
      // There doesn't seem to be any way out of the box to get the .mozilla 
      // directory (maybe there should be but there isn't).
      // (See https://developer.mozilla.org/index.php?title=en/Code_snippets/File_I%2F%2FO
      // http://mxr.mozilla.org/mozilla-central/source/xpcom/io/nsAppDirectoryServiceDefs.h
      // http://mxr.mozilla.org/mozilla-central/source/xpcom/io/nsDirectoryServiceDefs.h
      // .) But one can get the root profile direct of this app and take the 
      // parent directory
      this._profileDirectory = gDirectoryService.
                               get("DefProfRt", Ci.nsIFile).parent;

      // On platforms other than linux, the profile directory contains
      // an additional "Profiles" directory that we need to remove.  See
      // http://mxr.mozilla.org/mozilla-central/source/toolkit/xre/nsXREDirProvider.cpp#285
      if (gPlatform != "X11")
        this._profileDirectory = this._profileDirectory.parent;

      // append the app name
      this._profileDirectory.append(this._app);
    }
  },
  get app() {
    return this._app;
  },


  //
  // member functions
  //

  // initialize this object
  // @param app: the short name of the app ('firefox')
  // @param profDir: an nsIFile of the app's profile directory, or undefined
  //                 if you want ProfileManager to guess the directory
  // @param binary: an nsIFile of the default binary, or undefined if you
  //                want ProfileManager to use it's own default
  init: function(app, profDir, binary) {
    this.app = app;
    if (profDir) {
      this.profileDirectory = profDir;
    }

    // appconfig.ini is always in profilemanager's directory
    var appconfigIni = gDirectoryService.get("DefProfRt", Ci.nsIFile);
    if (gPlatform != "X11")
      appconfigIni = appconfigIni.parent;
    appconfigIni.append("appconfig.ini");
    this.appConfig = new appConfig(appconfigIni, 
                                   this.app, 
                                   this.appName,
                                   this.appProperties);

    // set the default app
    var app_default;
    if (binary) {
      app_default = binary.path;
    } else {
      app_default = this.appConfig.getDefaultApp();
      if (app_default == undefined) {
        app_default = findDefaultApp(this.app,
                                     this.appProperties);
      }
    }
    ProfileManagerService.defaultApp = app_default;

    // get the profiles.ini file
    var profilesIni = this._profileDirectory.clone();
    profilesIni.append('profiles.ini');
    // sanity check
    if (!profilesIni.exists()) {
      throw(profilesIni + " does not exist");
    }
    // instantiate the profile manager
    this.profileManager = new profileManager(profilesIni,
                                             this.appConfig);

    // if only one profile, make it the default
    if ((this.profileManager.profiles.length == 1) &&
        (!this.profileManager.profiles[0].Default)) {
      this.profileManager.setDefault(this.profileManager.profiles[0].name);
    }
  },
};
