/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Profile Manager.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Jonathan Griffin <jgriffin@mozilla.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// appconfig information is stored internally in the appconfig object,
// which has the following properties, where 'appname' is some application
// name, e.g., 'firefox':
//   appname: an array of objects, each with a 'path' and 'version' attribute
//   appname:default: an index into the appname array, which indicates the
//                    the default app version, or -1 if no default is set
//   appname:app: an object representing profile->app mappings,
//                where each object's key is a profile's name, and
//                the object's value is the corresponding app's path

function appConfig(appconfigIni, app, appname, appProperties) {
  this.appconfig = {};
  this.appconfigIni = appconfigIni;
  this.app = app;
  this.appname = appname;
  this.appProperties = appProperties;
  this.readIni();
}

appConfig.prototype.getDefaultApp = function() {
  // returns the default app, if it's been defined by the user
  let defaultIndex = this.appconfig[this.app + ":default"];
  if (typeof(defaultIndex) != "undefined" && defaultIndex != -1 &&
      typeof(this.appconfig[this.app][defaultIndex]) != "undefined") {
    return this.appconfig[this.app][defaultIndex];
  }
};

appConfig.prototype.writeIni = function() {
  // write the appconfig.ini file
  var foStream = Components.classes["@mozilla.org/network/file-output-stream;1"]
                 .createInstance(Components.interfaces.nsIFileOutputStream);

  // open the file with flags: PR_WRONLY | PR_CREATE_FILE | PR_TRUNCATE
  foStream.init(this.appconfigIni, 0x02 | 0x08 | 0x20, 0666, 0);
  var converter = Components.classes["@mozilla.org/intl/converter-output-stream;1"].
  createInstance(Components.interfaces.nsIConverterOutputStream);
  converter.init(foStream, "UTF-8", 0, 0);

  for (var section in this.appconfig) {
    if (section.indexOf(":default") > -1) {
      // no-op
    }
    else if (section.indexOf(":app") > -1) {
      // this is a profile->app mapping section; write verbatim
      converter.writeString("[" + section + "]\n");
      for (var key in this.appconfig[section]) {
        converter.writeString(key + "=" + this.appconfig[section][key] + "\n");
      }
      converter.writeString("\n");
    }
    else {
      // this is an app list; handle the default and non-default case
      converter.writeString("[" + section + "]\n");
      if (this.appconfig[section+":default"] == -1)
        converter.writeString("DEFAULT=\n");
      for (var i = 0; i < this.appconfig[section].length; i++) {
        if (i == this.appconfig[section+":default"])
          converter.writeString("DEFAULT=" + this.appconfig[section][i].path + "\n");
        else
          converter.writeString(this.appconfig[section][i].path + "=\n");
      }
      converter.writeString("\n");
    }
  }

  converter.close(); // this closes foStream
};

appConfig.prototype.readIni = function() {
  // read the ini file into the appconfig object
  if (!this.appconfigIni.exists())
    return;

  try {
    var parser = getINIParser(this.appconfigIni);
  } catch(e) {
    return;
  }
  let sectionEnumerator = parser.getSections();
  while (sectionEnumerator && sectionEnumerator.hasMore()) {
    let section = sectionEnumerator.getNext();
    let section_data = read_section(parser, section);
    if (section.indexOf(":app") > -1) {
      // This section contains profile->app mapping; just store as-is.
      this.appconfig[section] = section_data;
    }
    else {
      // This section contains a list of app paths.  Handle the default
      // case as well as populating app versions.  If a path does not exist,
      // don't add it to the list of app paths.
      this.appconfig[section] = [];
      this.appconfig[section+":default"] = -1;
      for (let key in section_data) {
        let appdata = {};
        if (key == "DEFAULT") {
          if (section_data[key]) {
            if (!fileExists(section_data[key])) {
              continue;
            }
            appdata['path'] = section_data[key];
            this.appconfig[section+":default"] = this.appconfig[section].length;
          }
          else {
            continue;
          }
        }
        else {
          if (!fileExists(key)) {
            continue;
          }
          appdata['path'] = key;
        }
        appdata['version'] = getAppVersion(appdata['path']);
        this.appconfig[section].push(appdata);
      }
    }
  }

  this.verifyProfileMappings();
};

appConfig.prototype.getAppList = function() {
  // returns an array of objects, each with 'path' and 'version' attributes,
  // comprising the list of app versions for this app
  return this.appconfig[this.app];
};

appConfig.prototype.updateAppList = function(defaultIndex, appList) {
  // update the app list in this.appconfig
  this.appconfig[this.app+":default"] = defaultIndex == undefined ? -1 :
    defaultIndex;
  this.appconfig[this.app] = [];
  let locatedDefaultPath = findDefaultApp(this.app, this.appProperties);
  if (locatedDefaultPath == undefined) {
    locatedDefaultPath = "none";
  }
  else {
    locatedDefaultPath = locatedDefaultPath.path;
  }
  for each (let app in appList) {
    let apppath = app;
    if (apppath['path']) {
      apppath = apppath.path;
    }
    this.appconfig[this.app].push({path: apppath, version: getAppVersion(apppath)});
    // If the automatically-located default app has been added to the list,
    // update the default index accordingly, in order to prevent an extra
    // copy of this app from appearing in appconfig.ini (and subsequently
    // in UI dialogs).
    if (this.appconfig[this.app+":default"] == -1 &&
        apppath.toLowerCase() == locatedDefaultPath.toLowerCase()) {
      this.appconfig[this.app+":default"] = this.appconfig[this.app].length - 1;
    }
  }
};

appConfig.prototype.verifyProfileMappings = function() {
  // Search through the list of profile->app mappings.  If a particular
  // mapping references an app that no longer exists in our list, 
  // delete it.
  for (let profile in this.appconfig[this.app + ":app"]) {
    let appPresent = false;
    for each (let appversion in this.appconfig[this.app]) {
      if (appversion.path.toLowerCase() == 
          this.appconfig[this.app + ":app"][profile].toLowerCase()) {
        appPresent = true;
      }
    }
    if (!appPresent) {
      delete this.appconfig[this.app + ":app"][profile];
    }
  }
};

appConfig.prototype.showAppVersionsDialog = function(window, defaultApp) {
  // show the dialog for managing app versions
  var params = { in: {appname: this.appname,
                      app: this.app,
                      apps: this.appconfig[this.app],
                      defaultIndex: this.appconfig[this.app+":default"],
                      defaultApp: defaultApp}, 
                 out: null };
  window.openDialog("chrome://profilemanager/content/appVersions.xul", 
                    "",
                    "modal", 
                    params);

  // update the app config from the dialog's output
  this.updateAppList(params.out.defaultIndex, params.out.apps);

  // verify that our profile->app mappings don't point to invalid apps
  this.verifyProfileMappings();

  // re-write the appconfig.ini file
  this.writeIni();

  // return the new default app, or null
  return this.appconfig[this.app+":default"] == -1 ? null : 
                                                     this.getDefaultApp();
};

appConfig.prototype.getAppForProfile = function(profileName, profilePath) {
  // get the app mapped to the profile with the specified name
  let profileApp = this.appconfig[this.app + ":app"];
  if (profileApp != undefined) {
    profileApp = profileApp[profilePath];
  }
  return profileApp == undefined ? 'Any' : profileApp;
};

appConfig.prototype.setAppForProfile = function(profileName, profilePath, appPath) {
  // set the app mapped to the given profile, and update the appconfig.ini file

  // if appPath is not in our app versions list, add it
  let needToAddApp = true;
  for each (let appVersion in this.appconfig[this.app]) {
    if (appPath && appVersion.path.toLowerCase() == appPath.toLowerCase()) {
      needToAddApp = false;
    }
  }
  if (appPath && needToAddApp) {
    if (this.appconfig[this.app] == undefined) {
      this.appconfig[this.app] = [];
    }
    this.appconfig[this.app].push({path: appPath, version: getAppVersion(appPath)});
  }

  // update the app list in order to remove any duplicates
  this.updateAppList(this.appconfig[this.app + ":default"], 
    this.appconfig[this.app].slice());

  // now update profile->app mapping
  if (this.appconfig[this.app + ":app"] == undefined) {
    this.appconfig[this.app + ":app"] = [];
  }
  if (appPath == undefined || appPath.toLowerCase() == "any") {
    if (this.appconfig[this.app + ":app"][profilePath]) {
      delete this.appconfig[this.app + ":app"][profilePath];
    }
  }
  else {
    this.appconfig[this.app + ":app"][profilePath] = appPath;
  }
  this.writeIni();
};

