/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 *
 * ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla Communicator client code, released
 * March 31, 1998.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2010
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Jeff Hammel <jhammel@mozilla.com>
 *    Mark Cote <mcote@mozilla.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

//// functions related to profiles

// service URIs
const ProcessService = "@mozilla.org/process/util;1";

// l10n string bundle
var gProfileManagerBundle = document.getElementById("bundle_profileManager");

//// functions for reading and writing profiles.ini

function profileManager(profilesIni, appConfig) {
  // nsIFile of profiles.ini
  this.profilesIni = profilesIni;
  // a copy of our appConfig instance
  this.appConfig = appConfig;

  // this is the relative path under profiles.ini, if any, where profiles
  // are stored
  this.profilesSubdirName = '';

  // we keep track of both the profiles.ini dir and the profiles dir,
  // since they may be different, and they get used in different ways
  this.profilesIniDirectory = this.profilesIni.parent;
  this.profilesDirectory = this.profilesIni.parent;

  // get platform
  var httpHandler = Components.classes["@mozilla.org/network/protocol;1?name=http"].
    getService(Components.interfaces.nsIHttpProtocolHandler);

  // on non-linux platforms, profiles are stored in a 'Profiles' subdir
  if (httpHandler.platform != "X11") {
    this.profilesSubdirName = 'Profiles';
    this.profilesDirectory.append(this.profilesSubdirName);
  }

  // get the directory for profile backups
  this.backupDir = this.profilesDirectory.clone();
  this.backupDir.append('backups');

  // this has to be added in the "constructor" since we call it therein
  this.readProfiles = function() {
    // reads the profiles from a given .ini file
    
    // as defined in    
    // http://mxr.mozilla.org/mozilla-central/source/toolkit/profile/public/nsIToolkitProfile.idl
    // profiles have the following attributes:
    
    // QueryInterface
    // localDir
    // lock
    // name
    // remove
    // rootDir  

    // reset the profile data
    this.profiles = new Array();
    this.general = {}; // the [General] section

    this.errors = {};  // an array per section that is appended to
    gProfileManagerBundle = document.getElementById("bundle_profileManager");

    // TODO: 
    // There should be a real profileError object that
    // can be intelligently used via e.g. ui.js

    // - orphaned backups (backups whose profiles are deleted) should be
    // displayed (somewhere) with the option of restoring from those backups
    
    var profileNames = {}; // to avoid profile name duplication

    // parse the ini file
    var iniParser = getINIParser(this.profilesIni);
    var sectEnum = iniParser.getSections();
    while (sectEnum && sectEnum.hasMore()) {
      var section = sectEnum.getNext();
      if (section == 'General') {
        this.general = read_section(iniParser, section);
      }
      else {
        // (try) to parse as a profile
        
        // basic data
        try {
          var name = iniParser.getString(section, 'Name');
          var path = iniParser.getString(section, 'Path');

          // check for duplicate names [TODO]
          if (profileNames[name]) {
            this.addError(name, gProfileManagerBundle.getFormattedString('profileSameNameError', [name]));
          }

        }
        catch(e) { 
          // error handling : should notify user and ask for actions
          this.addError('[' + section + ']', gProfileManagerBundle.getString('profileSectionReadError'));
          continue;
        }

        // relativity of path
        var isRelative = getBool(iniParser, section, 'IsRelative', true);
        // true == sensible default? ::shrug::
        
        // is the profile the default?
        var Default = getBool(iniParser, section, 'Default', false);
        
        // profile backups
        var Backups = this.backups(name);
        
        // directory of the profile
        var profileDirectory;
        if (isRelative) {
          // profiles paths are relative to the profiles.ini file, NOT
          // the profiles directory
          profileDirectory = this.profilesIniDirectory.clone();
          // relative paths can still contain path separators, which seem
          // to always be /, even on Windows
          let path_parts = path.split("/");
          for each (let path_part in path_parts)
            profileDirectory.append(path_part);
        }
        else {
          profileDirectory = getFile(path);
        }
        if (!profileDirectory.exists()) {
          // ensure profile exists
          this.addError(name, gProfileManagerBundle.getFormattedString('profilePathDoesNotExistError', [profileDirectory.path]));
          continue;
        }

        // read the profile
        var profile = read_section(iniParser, section);
        profile.section = section;
        profile.name = name;
        profile.path = path;
        profile.Default = Default;
        profile.Backups = Backups;
        profile.isRelative = isRelative;
        profile.rootDir = profileDirectory;

        // add the profile (if no errors)      
        this.profiles[this.profiles.length++] = profile;
      }
    }
  }

  // read the profiles
  this.readProfiles();
}

profileManager.prototype.isProfileLocked = function(profileName) {
  var profile = this.getProfile(profileName);

  if (gPlatform == "X11")
    var lockfile = ".parentlock";
  else if (gPlatform == "Macintosh")
    var lockfile = ".parentlock";
  else if (gPlatform == "Windows")
    var lockfile = "parent.lock";
  else
    // we don't attempt to detect lock status on other platforms
    return false;

  let check = profile.rootDir.clone();
  check.append(lockfile);

  if (gPlatform == "Macintosh" || gPlatform == "X11") {
    // first check for fcntl lock, then symlink lock
    if (isFcntlLocked(check)) {
      return true;
    }
    if (gPlatform == "X11") {
      // on linux, we use a different filename for symlink locks
      check = profile.rootDir.clone();
      check.append("lock");
    }
    try {
      if (check.isSymlink()) {
        return !isSymlinkStale(check);
      }
    }
    catch(e) {}
    return false;
  }
  else if (gPlatform == "Windows") {
    return check.exists();
  }

  return false;
}

profileManager.prototype.writeProfiles = function() {
  // writes the profiles to a given .ini file

  // as of this writing (Wed Jun 23 17:28:14 PDT 2010), 
  // internal-memory profiles are stored with the following attributes
  //
  // Backups: (list) backups  // TODO -> /dev/null
  // Default: (bool) whether profile is the default or not
  // IsRelative: [non-canonical]
  // Name: [non-canonical]
  // Path: [non-canonical]
  // isRelative: (bool) whether the profile is relative or absolute
  // name: (string) short name of the profile
  // path: (string) path of the profile
  // rootDir: (nsIFile) directory of the profile
  // section: section the profile is in; not needed here

  // if only one profile, set it as default
  if (this.profiles.length == 1) {
    this.profiles[0].Default = true;
  }

  // backup the existing profiles.ini (probably a good idea)
  this.profilesIni.copyTo(this.profilesIniDirectory, 'profiles.ini.bak');
  // TODO: error handling if this file exists

  // write the profiles.ini file
  var foStream = Components.classes["@mozilla.org/network/file-output-stream;1"]
                 .createInstance(Components.interfaces.nsIFileOutputStream);

  // use 0x02 | 0x10 to open file for appending.
  foStream.init(this.profilesIni, 0x02 | 0x08 | 0x20, 0666, 0);
  // write, create, truncate
  // In a c file operation, we have no need to set file mode with or operation,
  // directly using "r" or "w" usually.
  // if you are sure there will never ever be any non-ascii text in data you can
  // also call foStream.writeData directly
  var converter = Components.classes["@mozilla.org/intl/converter-output-stream;1"].
  createInstance(Components.interfaces.nsIConverterOutputStream);
  converter.init(foStream, "UTF-8", 0, 0);

  // write general section
  converter.writeString('[General]\n');
  for (var i in this.general) {
    converter.writeString(i + '=' + this.general[i] + '\n');
  }
  converter.writeString('\n');

  // write each profile
  for (var i in this.profiles) {
    var profile = this.profiles[i];
    converter.writeString('[Profile' + i + ']\n');
    converter.writeString('Name=' + profile.name + '\n');
    if (profile.isRelative) {
        var isRelative = 1;
      }
    else {
      var isRelative = 0;
    }
    converter.writeString('IsRelative=' + isRelative + '\n');
    converter.writeString('Path=' + profile.path + '\n');
    if (profile.Default) {
      converter.writeString('Default=1\n');
    }
    converter.writeString('\n');
  }

   converter.close(); // this closes foStream   

  // reread the profiles to ensure concurrency
  this.readProfiles();

  // remove backup of profiles.ini
  var backup = this.profilesIniDirectory.clone()
  backup.append('profiles.ini.bak');
  backup.remove(false);
}

// API functions

profileManager.prototype.newProfile = function(name, makeDefault, appPath, path, temporary) {
  // create a new profile
  // - name : name of the profile (extraneous if temporary)
  // - makeDefault : should the profile be the application default?
  // - appPath : path to the application to -CreateProfile with (string)
  // - path : path to the profile (string)
  // - temporary : is the profile temporary? don't add to profiles.ini

  if (!temporary) {
    // check name validity
    try {
      this.checkNewProfileName(name);
    } catch(e) {
      throw e;
    }
  }

  if (appPath && (!temporary)) {
    // TODO: deal with error conditions in the call to getFile
    var app = getFile(appPath);

    // make the profile with the app, if possible
    var process = Components.classes[ProcessService]
                  .createInstance(Components.interfaces.nsIProcess);

    // the profile name and directory should be quoted together when 
    // both are provided
    // see https://developer.mozilla.org/En/Command_Line_Options
    var profileArg = name;
    if (path) {
      profilArg += ' ' + path;
    }

    // use the app to create the profile
    process.init(app);
    var args = [ app.path,
                 '-CreateProfile',
                 profileArg,
                 '-no-remote' ] // yes, -no-remote needs to be used 
    process.run(true, args, 4);

    // reread the profiles
    this.readProfiles();

  } 
  else {
    // make a mock profile ourselves
    var relative = path ? false : true;
    if (path) {
      var profileDirectory = getFile(path);
      // XXX should probably check to see if this is an absolute path
      // and if not make it so
      
    } else {
      path = this.hash(name); // hashed directory name
      var profileDirectory = this.profilesDirectory.clone();
      profileDirectory.append(path);
    }

    // create the profile directory
    try {
      profileDirectory.create(profileDirectory.DIRECTORY_TYPE, 0700);
    } catch(e) {
      throw(e); 
    }

    if (temporary) { // we're done!
      return profileDirectory;
    }

    // make the profile the default if there is no other
    var isDefault = false;
    if (this.profiles.length == 0) {
      isDefault = true;
    }

    // add profile attributes
    var profile = { Backups: [],
                    Default: isDefault,
                    isRelative: relative,
                    name: name, 
                    path: relative ? (this.profilesSubdirName ? (this.profilesSubdirName + '/' + path) : path) : path,
                    rootDir: profileDirectory
    }
    
    // add the profile to memory
    this.profiles[this.profiles.length++] = profile;

    // write profiles.ini
    this.writeProfiles();
  }

  if (makeDefault) { 
    // set as default profile
    this.setDefault(name);
  }

  // return the profile data
  return this.getProfile(name);
}

profileManager.prototype.createTemporaryProfile = function() {
  // front end function to create a temporary profile
  // TODO: could pass the name of a profile to copy as a temporary
  
  // get the directoryService
  var DirectoryService = "@mozilla.org/file/directory_service;1";
  var directoryService = Components.classes[DirectoryService].
                         getService(Components.interfaces.nsIProperties);
  

  // get a temporary directory
  var temporaryDirectory = directoryService.get("TmpD", Components.interfaces.nsIFile);
  temporaryDirectory.append('profilemanager');
  if (!temporaryDirectory.exists()) {
    temporaryDirectory.create(temporaryDirectory.DIRECTORY_TYPE, 0700);
  }
  temporaryDirectory.append(Date.now());

  return this.newProfile(null, false, null, 
                         temporaryDirectory.path,
                         true);

}

profileManager.prototype.deleteProfile = function(name) {
  // delete the profile of given name

  // delete the profile from memory
  var index = this.profileIndex(name);
  if (index == undefined)
    throw(gProfileManagerBundle.getFormattedString('profileNotFoundError', [name]));

  var profile = this.profiles[index];

  // delete the profile from disk
  profile.rootDir.remove(true);

  this.profiles.splice(index, 1);

  // if we've deleted the default profile, make the first remaining profile
  // the default, otherwise our app might not know how to start
  if (profile.Default) {
    if (this.profiles.length > 0) {
      this.profiles[0].Default = true;
    }
  }

  // remove the profile from profiles.ini 
  this.writeProfiles();
}

profileManager.prototype.renameProfile = function(old_name, new_name) {
  // renames the profile of given name

  // no-op case
  if (old_name == new_name) { 
    return;
  }

  // ensure a legal name is used
  try {
    this.checkNewProfileName(new_name);
  } catch(e) {
    throw e; 
  }

  // save the app mapped to this profile, if any, so we can update
  // the mapping later
  var profile = this.getProfile(old_name);
  try {
    var oldProfileApp = this.appConfig.getAppForProfile(old_name, profile.path);
  }
  catch(e) {}

  // rename the profile in memory
  profile.name = new_name;

  // rename the profile directory if appropriate; skip renaming if the
  // directory doesn't contain the old profile's name
  var base = profile.rootDir.leafName;
  if (base.indexOf(old_name) > -1) {
    base = base.replace(old_name, new_name);
    var newRootDir = profile.rootDir.clone();
    try {
      newRootDir.moveTo(newRootDir.parent, base);
    }
    catch(e) {
      profile.name = old_name;
      throw(e);
    }
    profile.rootDir = newRootDir;
    if (profile.isRelative) {
      var parts = profile.path.split("/");
      var basePart = parts.pop();
      basePart = basePart.replace(old_name, new_name);
      parts.push(basePart);
      profile.path = parts.join("/");
    }
    else {
      profile.path = profile.rootDir.path;
    }
  }

  // rename the profile backups
  for (var backup in profile.Backups) {
    var file = profile.Backups[backup]['file'];
    var backupPath = file.leafName.split('.');
    backupPath[0] = new_name;
    backupPath = backupPath.join('.');
    file.moveTo(file.parent, backupPath);
  }

  // update app profile mappings for new name
  if (oldProfileApp && oldProfileApp != "Any") {
    this.appConfig.setAppForProfile(new_name, profile.path, oldProfileApp);
  }

  // update profiles.ini
  this.writeProfiles();
}

profileManager.prototype.backupToArchive = function(archiveFile, profileName, callback) {
  /* export a profile to a .zip archive
   * @param archiveFile - an nsIFile of the archive to create
   * @param profileName - the name of the profile to archive
   * @param callback - a callback that this function will call to notify
   *       the caller of progress.  The callback is called as
   *       function(hasMore), where hasMore is true or false, depending
   *       on whether there is more data to archive.  The function will
   *       always be called with hasMore = false at the end of the operation.
   *       The callback should return true or false, depending on whether
   *       it wishes to continue the archiving or not.
   */

  // list of files to be added to the archive asynchronously
  var filelist = [];

  function addDirToArchive(zipW, currentDir, relativePath) {
    // adds a directory to an existing zipfile object
    // params:
    //   zipW - an existing nsIZipWriter instance
    //   currentDir - an nsIFile directory of files to add
    //   relativePath - the relativePath of this directory inside the zipfile,
    //       must use forward slashes according to 
    //       https://developer.mozilla.org/en/nsIZipWriter#addEntryFile.28.29
    var filesEnum = currentDir.directoryEntries;
    while (filesEnum.hasMoreElements()) {
      let file = filesEnum.getNext().QueryInterface(Components.interfaces.nsILocalFile);
      if (file.isDirectory()) {
        zipW.addEntryDirectory(relativePath + file.leafName, Date.now() * 1000, false);
        addDirToArchive(zipW, file, relativePath + file.leafName + "/");
      }
      else {
        let filepath = relativePath + file.leafName;
        // we don't add files synchronously; instead we add them to an array
        // and add them to the archive later, asynchronously, so that the
        // progress meter can continue to animate
        filelist.push({path: filepath, file: file});
      }
    }
  }

  var profile = this.getProfile(profileName);
  var zipWriter = Components.Constructor("@mozilla.org/zipwriter;1", "nsIZipWriter");
  var zipW = new zipWriter();
  // PR_RDWR | PR_CREATE_FILE | PR_TRUNCATE
  try {
    zipW.open(archiveFile, 0x04 | 0x08 | 0x20);
  }
  catch(e) {
    alert(gProfileManagerBundle.getFormattedString('profileCannotCreateZipError', [archiveFile.path, e]));
  }

  function isZipComplete() {
    if (filelist.length == 0) {
      zipW.close();
      setTimeout(function() { doCallback(false); }, 100);
    }
    else {
      let afile = filelist.pop();
      try {
        // exceptions can be thrown if the file can't be opened (e.g., 
        // parent.lock for in-use profiles)
        zipW.addEntryFile(afile.path, 
          Components.interfaces.nsIZipWriter.COMPRESSION_DEFAULT,
          afile.file,
          false);
      }
      catch(e) { /* no-op */ }
      setTimeout(function() { doCallback(true); }, 0);
    }
  }

  function doCallback(hasMore) {
    let shouldContinue = callback.call(this, hasMore);
    if (shouldContinue && hasMore) {
      isZipComplete();
    }
    else if (!shouldContinue && hasMore) {
      filelist = [];
      isZipComplete();
      if (archiveFile.exists())
        archiveFile.remove(false);
    }
    else { /* no-op */ }
  }

  addDirToArchive(zipW, profile.rootDir, "");
  setTimeout(function() { doCallback(true); }, 100);
}


profileManager.prototype.createFromArchive = function(archiveFile, profileName, callback) {
  /* creates a profile from a .zip archive
   * @param archiveFile - an nsIFile of the archive to open
   * @param profileName - the name of the profile to create
   * @param callback - a callback that this function will call to notify
   *       the caller of progress.  The callback is called as
   *       function(hasMore), where hasMore is true or false, depending
   *       on whether there is more data to extract.  The function will
   *       always be called with hasMore = false at the end of the operation.
   *       The callback should return true or false, depending on whether
   *       it wishes to continue the extraction or not.
   */

  // create the new profile's object and initialize some properties
  var newProfile = {};
  newProfile['name'] = profileName;
  newProfile['Default'] = false;
  newProfile['isRelative'] = true;
  newProfile['path'] = this.hash(profileName);

  var rootDir = this.profilesDirectory.clone();
  rootDir.append(newProfile['path']);
  newProfile['rootDir'] = rootDir;
  if (this.profilesSubdirName) {
    // update relative path so it's relative to profiles.ini
    newProfile['path'] = this.profilesSubdirName + '/' + newProfile['path'];
  }

  function getItemFile(filePath) {
    // returns an nsIFile for rootDir + filePath
    var itemLocation = rootDir.clone();
    var parts = filePath.split("/");
    for (let i = 0; i < parts.length; i++) {
      itemLocation.append(parts[i]);
    }
    return itemLocation;
  }

  // open the archive
  var zipReader = Components.classes["@mozilla.org/libjar/zip-reader;1"]
                            .createInstance(Components.interfaces.nsIZipReader);
  zipReader.open(archiveFile);

  // create directories first
  var entries = zipReader.findEntries("*/");
  var fileEntries = zipReader.findEntries(null);
  var that = this;
  
  function finishExtracting(canceled) {
    zipReader.close();
    if (canceled) {
      if (newProfile['rootDir'].exists())
        newProfile['rootDir'].remove(true);
    }
    else {
      // add the new profile to this.profiles
      that.profiles.push(newProfile);

      // update profiles.ini and UI
      that.writeProfiles();
    }
    setTimeout(function() { doCallback(false); }, 0);
  }

  function continueExtracting() {
    if (entries.hasMore()) {
      var entryName = entries.getNext();
      var target = getItemFile(entryName);
      if (!target.exists()) {
        try {
          target.create(Components.interfaces.nsILocalFile.DIRECTORY_TYPE, 0777);
        }
        catch (e) {
          alert(gProfileManagerBundle.getFormattedString('profileCannotCreateTargetDirError', [target.path, e]) + "\n");
        }
      }
      setTimeout(function() { doCallback(true); }, 0);
    }
    else if (fileEntries.hasMore()) {
      var entryName = fileEntries.getNext();
      target = getItemFile(entryName);
      if (!target.exists()) {
        try {
          target.create(Components.interfaces.nsILocalFile.DIRECTORY_TYPE, 0777);
        }
        catch (e) {
          alert(gProfileManagerBundle.getFormattedString('profileCannotCreateTargetFileError', [target.path, e]) + "\n");
        }
        zipReader.extract(entryName, target);
      }
      setTimeout(function() { doCallback(true); }, 0);
    }
    else {
      finishExtracting(false);
    }
  }

  function doCallback(hasMore) {
    let shouldContinue = callback.call(this, hasMore);
    if (shouldContinue && hasMore) {
      continueExtracting();
    }
    else if (!shouldContinue && hasMore) {
      finishExtracting(true);
    }
    else { /* no-op */ }
  }

  setTimeout(function() { doCallback(true); }, 100);
}

/**
 * Creates a copy of a profile.
 * @param name The name of the profile to copy
 * @param clone_name The new name for the profile copy
 * @param path {string} The path for the profile copy. If undefined, the
 *        profile copy will be given a path in the app's profile directory.
 */
profileManager.prototype.cloneProfile = function(name, clone_name, path) {
  // check new profile name
  try {
    this.checkNewProfileName(clone_name);
  } catch(e) {
    throw e;
  }

  // get the original profile
  var profile = this.getProfile(name);
  var newProfile = new Object();
  newProfile['name'] = clone_name;
  newProfile['Default'] = false;

  // copy the directory 
  if (path) {
    var profileDir = getFile(path);
    newProfile['isRelative'] = false;
    newProfile['path'] = profileDir.path;

    // this may not work, see
    // http://mxr.mozilla.org/mozilla-central/source/xpcom/io/nsIFile.idl#120
    profile.rootDir.copyTo(profileDir.parent, profileDir.leafName);

    newProfile['rootDir'] = profileDir;

  } else {
    newProfile['isRelative'] = true;
    newProfile['path'] = this.hash(clone_name);
    profile.rootDir.copyTo(this.profilesDirectory, newProfile['path']);
    var rootDir = this.profilesDirectory.clone();
    rootDir.append(newProfile['path']);
    newProfile['rootDir'] = rootDir;
    if (this.profilesSubdirName) {
      // update relative path so it's relative to profiles.ini
      newProfile['path'] = this.profilesSubdirName + '/' + newProfile['path'];
    }
  }

  // add the new profile to this.profiles
  this.profiles.push(newProfile);

  // update profiles.ini
  this.writeProfiles();
}

profileManager.prototype.setDefault = function(name) {
  // set the given profile as the default

  // ensure the given profile exists
  var profile = this.getProfile(name); 
  
  // set profile as the default, set others as not
  for (var i in this.profiles) {
    profile = this.profiles[i];
    if (profile.name == name) { 
      profile.Default = true;
    }
    else {
      profile.Default = false;
    }
  }

  // update profiles.ini
  this.writeProfiles();
}

profileManager.prototype.defaultProfile = function() {
  // returns the default profile, if any

  for (var i in this.profiles) {
    if (this.profiles[i].Default) {
      return this.profiles[i];
    }
  }
}

profileManager.prototype.deleteBackup = function(name, path) {
  // delete a backup profile with the given path, return true if successful
  var profile = this.getProfile(name);

  // make sure backup list is current
  profile.Backups = this.backups(profile.name);

  // find the backup with the specified path
  for (var i = 0; i < profile.Backups.length; i++) {
    if (profile.Backups[i].file.path.toLowerCase() == path.toLowerCase()) {
      var file = getFile(path);
      if (file.exists()) {
        file.remove(true);
        // rebuild the profile list
        profile.Backups = this.backups(name);
        return true;
      }
    }
  }
  return false;
}

profileManager.prototype.backupProfile = function(name) {
  // generate a backup of the given profile

  // get the profile
  var profile = this.getProfile(name);

  // create the backup directory if it doesn't exist
  if (!this.backupDir.exists()) {
    this.backupDir.create(this.backupDir.DIRECTORY_TYPE, 0700);
  }
  
  // make a backup directory for the profile: name.timestamp.bak
  var datestamp = Date.now();
  var backupdir = this.backupDir.clone();

  // copy the profile to the backup directory
  var backupName = profile.name + '.' + datestamp + '.bak'
  try {
    profile.rootDir.copyTo(backupdir, backupName);
  }
  catch(e) {
    // cleanup
    backupdir.append(backupName);
    backupdir.remove(true);
    throw(e);
  }

  // repopulate backups list [could just prepend instead]
  profile.Backups = this.backups(profile.name);
}

profileManager.prototype.restoreProfile = function(name, date, to_delete) {
  // restore the profile from a given date

  // get the profile
  var profile = this.getProfile(name);

  // find the requested backup
  var backups = this.backups(name);
  if (!backups.length) { // ensure backups exist for the profile
    throw gProfileManagerBundle.getFormattedString('profileNoBackupsFoundError', [name, this.backupDir.path]);
  }
  if (date) {
    // restore from the backup one older than or equal to the date 
    var found = false;
    for (var i in backups) {
      if (backups[i].date <= date) {
        var backup = backups[i];
        found = true;
        break;
      }
    }
    
    if (!found) { 
      throw gProfileManagerBundle.getFormattedString('profileNoOlderBackupFoundError', [date]);
    }    
  }
  else {
    // restore from the most recent backup
    var backup = backups[0];
  }

  // make a temporary copy of the profile in case seomthing goes wrong
  var backupdir = this.backupDir.clone();
  var tmpProfileBackup = this.backupDir.clone();
  tmpProfileBackup.append(name + '.tmp');
  try {
    profile.rootDir.copyTo(backupdir, profile.name + '.tmp');
  }
  catch(e) {
    // cleanup
    tmpProfileBackup.remove(true);
    throw(e);
  }

  // copy the backup over the profile
  var profiledir = profile.rootDir.clone();
  var path = profiledir.leafName;
  try {
    profiledir.remove(true);
    backup['file'].copyTo(profiledir.parent, path);

    // delete the backup if specified
    if (to_delete) {
      backup['file'].remove(true);
      profile.Backups = this.backups(profile.name);
    }

  }
  catch(e) {
    // resore profile from temporary copy
    tmpProfileBackup.copyTo(profiledir.parent, path);
    throw e;
  }

  // delete temporary copy of the profile
  if (tmpProfileBackup.exists())
    tmpProfileBackup.remove(true);
}

profileManager.prototype.backups = function(name) {
  // reverse chronological list of backups for a given profile

  var backups = [];
  if (!this.backupDir.exists()) {
    return backups;
  }

  // find backups for the given profile
  var directoryEntries = this.backupDir.directoryEntries;
  while (directoryEntries.hasMoreElements()) {
    var entry = directoryEntries.getNext().QueryInterface(Components.interfaces.nsIFile);
    var filename = entry.leafName;

    // the filename should be in three parts:
    // (profile).(date).bak
    // - profile : the name of the profile
    // - date : milliseconds since epoch ( see 
    //   http://www.w3schools.com/js/js_obj_date.asp )
    var parts = filename.split('.');

    if (parts.length != 3) { 
      continue;
    }
    if (parts[0] != name) {
      continue;
    }
    
    // (try to) convert the date part to an integer
    try {
      var date = parseInt(parts[1]);
    } catch(e) {
      continue;
    }

    // create the backup
    var backup = {};
    backup['file'] = entry;
    backup['date'] = new Date(date);

    // append the backup to the array
    backups[backups.length++] = backup;
  }

  // sort backups by reverse date
  function sort_backups(a, b) {
    return a.date < b.date;
  }
  backups.sort(sort_backups);

  return backups;
}

// utility functions

profileManager.prototype.profileIndex = function(name) {
  // returns the index of a profile with a given name
  // XXX should probably use some mapping functions for faster lookup
  for (var i in this.profiles) {
    if (this.profiles[i].name == name) {
      return i;
    }
  }
}

profileManager.prototype.getProfile = function(name) {
  // get a Profile given its name
  var index = this.profileIndex(name);
  if (index == undefined) { 
    throw(gProfileManagerBundle.getFormattedString('profileNotFoundError', [name]));
  }
  return this.profiles[index];
}
  
profileManager.prototype.hash = function(name) {
  // hashed directory name

  // possible characters for the profile hash
  var population = 'abcdefghijklmnopqrstuvwxyz0123456789';
    
  var hash = '';
  for (var i=0; i<8; i++) { 
    var index = Math.floor(Math.random()*population.length);
    hash += population[index];
  }
  return hash + '.' + name;
  
}

profileManager.prototype.checkNewProfileName = function(name) {
  // checks the validity of a new profile name
  // throws an exception if the name is invalid, 
  // otherwise returns true

  // name shouldn't be the empty string
  if (!name.length) {
    throw gProfileManagerBundle.getString('profileEmptyNameError');
  }

  // name should have no space or slashes in it
  var illegal_chars = ' /\\.';
  var found_illegals = '';
  for (var i in illegal_chars) { 
    var c = illegal_chars[i];
    if (name.indexOf(c) != -1) {
      found_illegals += c;
    }
  }
  if (found_illegals.length) {
    throw gProfileManagerBundle.getFormattedString('profileIllegalNameError', [found_illegals]);

  }

  // - name should not already be a profile
  if (this.profileIndex(name) != undefined) {
    throw gProfileManagerBundle.getFormattedString('profileAlreadyExistsError', [name]);
  }

  return true; // name is fine
}

profileManager.prototype.addError = function(name, error) {
  // add a profile-reading error

  if (!this.errors[name]) {
    this.errors[name] = new Array();
  }
  this.errors[name].push(error);
}
